(*==========================================================================*)
(* Muster-Compiler fuer eine PASCAL-aehnliche Trivialsprache                *)
(*                                                                          *)
(* Semantische Analyse und Codeerzeugung                                    *)
(*                                                                          *)
(*==========================================================================*)

unit CODER;

interface

  uses SCANNER;

  var PROGRAMMNAME: STRING8;
      CODEFILE    : TEXT;

  procedure UPN(SYMBL: SYMBOL; ATTR: INTEGER);

implementation

type
     STRING20 = STRING[20];
     STRING80 = STRING[80];
                                 (* Operand im Assemblerbefehl kann sein:   *)
     KLASSE   = (KONST,          (* Konstante                               *)
                 VARIAB,         (* Variable                                *)
                 FELD,           (* Feldadresse                             *)
                 ADR_REG,        (* Adresse eines Feldelementes im Register *)
                 REGISTER);      (* Zwischenergebnis im Register            *)

     PC_REGISTER    = (AX,BX,CX,SI,DI,DX);
     REGISTERSATZ   = set of PC_REGISTER;

var  FREIE_REGISTER   : REGISTERSATZ; (* Menge der freien Register          *)
     MARKENZAEHLER    : INTEGER;
     DEKLARATIONSTEIL : BOOLEAN;      (* fuer korrekte Behandlung von Namen *)
     NB_INDEX         : INTEGER;      (* Index eines Namens im Namensbuch   *)

const
                                      (* Registerbezeichnungen
                                         (als Wort- oder Byte-Register)
                                         'XX' bedeutet: kein Byte-Register  *)
     REGISTER_NAME    : ARRAY [INTEGER_SYM..BOOLEAN_SYM,
                               PC_REGISTER] OF STRING[2] =
                        (('AX','BX','CX','SI','DI','DX'), (* 16bit-Register *)
                         ('AL','BL','CL','XX','XX','XX'));(* 8bit-Register  *)

                                      (* Type-Override fr Assembler *)
     TYPE_            : ARRAY [INTEGER_SYM..BOOLEAN_SYM] OF STRING[9] =
                        ('Word Ptr ','Byte Ptr ');

type
                       (* jeder Operand hat einen Typ (z.B. INTEGER),
                          eine Klasse (z.B. KONSTANTE) und eine
                          Identifikation (z.B. Konstantenwert) *)
     OPERAND = record
                       TYP    : SYMBOL;            (* INTEGER oder BOOLEAN *)
                       ART    : KLASSE;            (* Art des Operanden    *)
                       case BOOLEAN of
                         TRUE: (NUMMER : INTEGER); (* Konstantenwert oder  *)
                                                   (* Variablennummer      *)
                         FALSE:(REG : PC_REGISTER) (* Registernummer       *)
               end;

const
                        (* vordefinerte Operanden: *)

     EINS: OPERAND = (TYP:INTEGER_SYM;ART:KONST;NUMMER:1);
     REG_AX  : OPERAND = (TYP:INTEGER_SYM;ART:REGISTER;REG:AX);

var
     KELLER: array[1..20] of OPERAND;     (* Operandenkeller                *)
     TOP   : 0..20;                       (* Zeiger auf oberstes Element    *)

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(*procedure UPN;*)                     (* SEMANTISCHE ANALYSE UND CODEERZEUGUNG *)
procedure UPN(SYMBL: SYMBOL; ATTR: INTEGER);
(*--------------------------------------------------------------------------*)
(*            Meldung eines semantischen Fehlers                            *)
(*--------------------------------------------------------------------------*)

procedure SEM_ERROR(MELDUNG: STRING80);
  begin
      WRITELN(' !! ',MELDUNG,' !!');
      HALT;
  end;

procedure TYP_FEHLER(WO: STRING80);
  begin
    SEM_ERROR(CONCAT('Fehlerhafter Typ bei ',WO));
  end;

(*--------------------------------------------------------------------------*)
(*            Registerverwaltung                                            *)
(*--------------------------------------------------------------------------*)

procedure REGISTER_BELEGEN(RNUM: PC_REGISTER);
  begin
    FREIE_REGISTER:=FREIE_REGISTER-[RNUM];
  end;

(*--------------------------------------------------------------------------*)

procedure REGISTER_FREIGEBEN(RNUM: PC_REGISTER);
  begin
    FREIE_REGISTER:=FREIE_REGISTER+[RNUM];
  end;

(*--------------------------------------------------------------------------*)
(*          freies Register bestimmen und belegen                           *)
(*--------------------------------------------------------------------------*)

function FREIES_REGISTER: PC_REGISTER;
  var I: PC_REGISTER;

  (* Achtung: DX wird nie als freies Register ermittelt, da es fuer
     Multiplikation und Division benoetigt wird *)

  begin
    if FREIE_REGISTER=[] then                 (* wenn alle Register belegt  *)
      begin
        SEM_ERROR('Ausdruck zu kompliziert !!');
      end;
    I:=AX;                                    (* mit AX anfangen            *)
    while not (I in FREIE_REGISTER) do
      I:=SUCC(I);
    FREIE_REGISTER:=FREIE_REGISTER-[I];       (* gefundenes Register aus    *)
    FREIES_REGISTER:=I;                       (* Menge der freien austragen *)
  end;

function FREIES_INDEX_REGISTER: PC_REGISTER;
  var I: PC_REGISTER;
  begin
    if (FREIE_REGISTER * [SI,DI])=[] then    (* wenn kein Indexregister frei*)
      begin
        SEM_ERROR('Ausdruck zu kompliziert !!');
      end;
    I:=SI;
    while not (I in FREIE_REGISTER) do
      I:=SUCC(I);
    FREIE_REGISTER:=FREIE_REGISTER-[I];
    FREIES_INDEX_REGISTER:=I;
  end;

(*--------------------------------------------------------------------------*)
(*            Kellerverwaltung                                              *)
(*--------------------------------------------------------------------------*)

procedure PUSH(OBJEKT: OPERAND);
  begin
    if TOP=20 then
      SEM_ERROR('Kellerueberlauf !!')
    else
      begin
        TOP:=TOP+1;
        KELLER[TOP]:=OBJEKT;
        (* etwaiges Arbeitsregister belegen *)
        if OBJEKT.ART in [REGISTER,ADR_REG] then
          REGISTER_BELEGEN(OBJEKT.REG);
      end;
  end;

(*--------------------------------------------------------------------------*)

procedure POP(var OBJEKT: OPERAND);
  begin
    if TOP=0 then
      SEM_ERROR('Kellerunterlauf !!')
    else
      begin
        OBJEKT:=KELLER[TOP];
        TOP:=TOP-1;
        if OBJEKT.ART in [REGISTER,ADR_REG] then
          REGISTER_FREIGEBEN(OBJEKT.REG);
      end;
  end;

(*--------------------------------------------------------------------------*)
(*            Umwandlung Kellerelement -> Assemblernotation                 *)
(*--------------------------------------------------------------------------*)

function ASSEMBLERNOTATION(OBJEKT: OPERAND): STRING20;
  var WERT: STRING[2];
  begin
    if OBJEKT.ART=FELD then                   (* Ein Feldname darf niemals  *)
      SEM_ERROR('Falscher Feldzugriff');      (* unmittelbarer Operand sein *)
    case OBJEKT.ART of
      KONST      : begin                      (* z.B.: 3765                 *)
                     STR(OBJEKT.NUMMER,WERT); (* Wandlung Zahl -> String    *)
                     ASSEMBLERNOTATION:=WERT;
                   end;
      VARIAB     : with NB[OBJEKT.NUMMER] do  (* z.B.: Word Ptr [_PAUL]     *)
                     ASSEMBLERNOTATION:=TYPE_[TYP]+'[_'+NAME+']';
      ADR_REG    :                            (* z.B.: Word Ptr [SI]        *)
                   ASSEMBLERNOTATION:=TYPE_[OBJEKT.TYP]+'['+
                                      REGISTER_NAME[INTEGER_SYM,OBJEKT.REG]+
                                      ']';
      REGISTER   : begin                      (* z.B.: BX                   *)
                     WERT:=REGISTER_NAME[OBJEKT.TYP,OBJEKT.REG];
                     ASSEMBLERNOTATION:=WERT;
                     if WERT='XX' then
                       SEM_ERROR('Ausdruck zu kompliziert');
                   end;
    end;
  end;

(*--------------------------------------------------------------------------*)
(*            Erzeugung eines Assemblerbefehls                              *)
(*--------------------------------------------------------------------------*)

procedure BEFEHL(OPCODE: STRING8; OPND_1,OPND_2: OPERAND);
  begin
    WRITELN(CODEFILE,OPCODE:12+LENGTH(OPCODE),' ':5-LENGTH(OPCODE),
            ASSEMBLERNOTATION(OPND_1),',',
            ASSEMBLERNOTATION(OPND_2));
  end;

(*--------------------------------------------------------------------------*)
(*            Operand in ein Arbeitsregister bringen                        *)
(*--------------------------------------------------------------------------*)

procedure LOAD(var OPND: OPERAND);
  var ZIEL: OPERAND;
      LADEBEFEHL: STRING8;
  begin
    with OPND do
      if ART<>REGISTER then            (* wenn nicht bereits im Register    *)
        begin
          ZIEL.ART:=REGISTER;
          ZIEL.TYP:=TYP;
          if ART=ADR_REG then          (* wenn OperandenADRESSE im Register *)
            REGISTER_FREIGEBEN(REG);
          ZIEL.REG:=FREIES_REGISTER;   (* neues freies Register ermitteln   *)
          BEFEHL('MOV',ZIEL,OPND);
          OPND:=ZIEL;
        end
  end;

(*--------------------------------------------------------------------------*)
(*            Bearbeitung fuer +,-,*,/,OR,AND                               *)
(*--------------------------------------------------------------------------*)

procedure DYADISCHER_OPERATOR(TEST_TYP: SYMBOL);
                                    (* TEST_TYP = erwarteter Operanden-Typ  *)
  var NEU,LOP,ROP: OPERAND;
      i: INTEGER;
  begin
    if not(AX IN FREIE_REGISTER) and (* wenn AX belegt ist, dann muss es bei*)
       (SYMBL in [MAL,GETEILT]) then (* Multiplikat. und Division gerettet  *)
      begin                          (* werden, da es dort benoetigt wird   *)
        NEU.REG:=FREIES_REGISTER;
        NEU.ART:=REGISTER;
        NEU.TYP:=INTEGER_SYM;
        BEFEHL('MOV',NEU,REG_AX);     (* AX in freies Register umspeichern  *)
                                      (* und entspr. Kellereintrag aendern  *)
        for i:=1 to TOP do with KELLER[I] do
          if (ART in [REGISTER,ADR_REG]) and (REG=AX) then
            NUMMER:=NEU.NUMMER;
        FREIE_REGISTER:=FREIE_REGISTER+[AX];
      end;
    (* Ladebefehl muss erzeugt werden, bevor Operanden aus Stack geholt
       werden, damit etwaige belegte Register nicht vorzeitig durch POP
       freigegeben werden
    *)
    LOAD(KELLER[TOP-1]);                             (*  MOVE  reg,LOP   *)
    POP(ROP);
    POP(LOP);
    if not( (LOP.TYP = TEST_TYP) and (ROP.TYP = TEST_TYP) ) then
      TYP_FEHLER('dyad. Operator');
    case SYMBL of
      AND_SYM   : BEFEHL('AND',LOP,ROP);             (*  AND   reg,ROP   *)
      OR_SYM    : BEFEHL('OR',LOP,ROP);              (*  OR    reg,ROP   *)
      PLUS      : BEFEHL('ADD',LOP,ROP);             (*  ADD   reg,ROP   *)
      MINUS     : BEFEHL('SUB',LOP,ROP);             (*  SUB   reg,ROP   *)
      MAL       : WRITELN(CODEFILE,'IMUL ':17,
                          ASSEMBLERNOTATION(ROP));   (*  IMUL  ROP       *)
      GETEILT   : begin
                    (* Dividend muss auf 32 bit erweitert werden *)
                    WRITELN(CODEFILE,'CWD':15);      (*  CWD             *)
                    WRITELN(CODEFILE,'IDIV ':17,
                          ASSEMBLERNOTATION(ROP));   (*  IDIV  ROP       *)
                  end;
    end;
    PUSH(LOP);               (* Ergebnis-Operand *)
  end;

(*--------------------------------------------------------------------------*)
(*            Bearbeitung fuer >,<,=                                        *)
(*--------------------------------------------------------------------------*)

procedure VERGLEICHSOPERATOR;
  var LOP,ROP: OPERAND;
  begin
    LOAD(KELLER[TOP-1]);                             (*  MOV   reg,LOP   *)
    POP(ROP);
    POP(LOP);
    if (LOP.TYP<>ROP.TYP) or
       ((LOP.TYP=BOOLEAN_SYM) and not (SYMBL in [GLEICH,UNGLEICH])) then
      TYP_FEHLER('Vergleichsoperator');
    BEFEHL('CMP',LOP,ROP);                           (*  CMP   LOP,ROP   *)
    WRITE(CODEFILE,'SET':15);                        (*  SETcc LOP       *)
    case SYMBL of
      GLEICH    : WRITE(CODEFILE,'E');
      UNGLEICH  : WRITE(CODEFILE,'NE');
      GROESSER  : WRITE(CODEFILE,'G');
      KLEINER   : WRITE(CODEFILE,'L')
    end;
    (* jetzt wird geprft, ob der linke Operand in einem Register steht,
       das NICHT fr Byte-Operanden benutzt werden kann, z.B. SI. In
       diesem Fall wird dieses register frei gegeben und ein anderes
       geeignetes Register fr das Vergleichsergebnis gesucht
    *)
    if REGISTER_NAME[BOOLEAN_SYM,LOP.REG]='XX' then
      begin
        REGISTER_FREIGEBEN(LOP.REG);
        LOP.REG:=FREIES_REGISTER;
        if REGISTER_NAME[BOOLEAN_SYM,LOP.REG]='XX' then
          SEM_ERROR('Ausdruck zu kompliziert !!');
      end;
    WRITELN(CODEFILE,' ',REGISTER_NAME[BOOLEAN_SYM,LOP.REG]);
    LOP.TYP:=BOOLEAN_SYM;
    PUSH(LOP);
  end;

(*--------------------------------------------------------------------------*)
(*            Bearbeitung fuer :=                                           *)
(*--------------------------------------------------------------------------*)

procedure ZUW_OPERATOR;
  var LOP,ROP: OPERAND;
  begin
    LOAD(KELLER[TOP]);                               (*  MOV   reg,ROP   *)
    POP(ROP);
    POP(LOP);
    if LOP.TYP<>ROP.TYP then
      TYP_FEHLER(':=');
    BEFEHL('MOV',LOP,ROP);                           (*  MOV   LOP,reg   *)
  end;

(*--------------------------------------------------------------------------*)
(*            Bearbeitung fuer IF,THEN,ELSE,END_IF                          *)
(*--------------------------------------------------------------------------*)

procedure IF_BEARBEITUNG;
  var BEDINGUNG,MARKE: OPERAND;
  begin
    case SYMBL of
      IF_SYM:
            begin
              MARKE.NUMMER:=MARKENZAEHLER;
              PUSH(MARKE);
              MARKENZAEHLER:=MARKENZAEHLER+2;
            end;
      THEN_SYM:
            begin
              LOAD(KELLER[TOP]);      (* IF-Bedingung in Register laden     *)
              POP(BEDINGUNG);
              if BEDINGUNG.TYP <> BOOLEAN_SYM then
                TYP_FEHLER('IF oder DO');
              BEFEHL('CMP ',BEDINGUNG,EINS);
              WRITELN(CODEFILE,'JNE  _':18,          (*   JNE     _j    *)
                KELLER[TOP].NUMMER);
            end;                                     (*     .....       *)
      ELSE_SYM:
            begin
              WRITELN(CODEFILE,'JMP  _':18,          (*   JMP    _j+1   *)
                KELLER[TOP].NUMMER+1);
              WRITELN(CODEFILE,'_',
                KELLER[TOP].NUMMER,':');             (* _j              *)
              KELLER[TOP].NUMMER:=KELLER[TOP].NUMMER+1;
            end;                                     (*     .....       *)
      END_IF:
            begin
              POP(MARKE);
              WRITELN(CODEFILE,'_',MARKE.NUMMER,':');(* _j+1            *)
            end
    end
  end;

(*--------------------------------------------------------------------------*)
(*            Bearbeitung fuer WHILE,DO,END_WHILE                           *)
(*--------------------------------------------------------------------------*)

procedure WHILE_BEARBEITUNG;
  var MARKE,BEDINGUNG: OPERAND;
  begin
    case SYMBL of
      WHILE_SYM:
            begin
              MARKE.NUMMER:=MARKENZAEHLER;
              PUSH(MARKE);
              WRITELN(CODEFILE,'_',MARKE.NUMMER+1,':');  (* _j+1            *)
              MARKENZAEHLER:=MARKENZAEHLER+2;
            end;                                         (*     .....       *)
      DO_SYM:
            begin
              SYMBL:=THEN_SYM;
              IF_BEARBEITUNG;          (* DO erzeugt gleichen Code wie THEN *)
            end;
      END_WHILE:
            begin
              POP(MARKE);
              WRITELN(CODEFILE,'JMP  _':18,              (*   JMP    _j+1   *)
                MARKE.NUMMER+1);
              WRITELN(CODEFILE,'_',MARKE.NUMMER,':');    (* _j              *)
            end
    end
  end;

(*--------------------------------------------------------------------------*)
(*            Bearbeitung fuer monadisches Minus                            *)
(*--------------------------------------------------------------------------*)

procedure KOMPLEMENTIEREN;
  begin
    with KELLER[TOP] do
      begin
        if ART=KONST then            (* Konstante im Keller komplementieren *)
          NUMMER:=-NUMMER
        else
          begin
            LOAD(KELLER[TOP]);       (* Operand in Register laden           *)
            if TYP <> INTEGER_SYM then
              TYP_FEHLER('-');
            WRITELN(CODEFILE,'NEG  ':17,ASSEMBLERNOTATION(KELLER[TOP]));
          end;
      end;
  end;

(*--------------------------------------------------------------------------*)
(*            Bearbeitung fuer Zugriff auf Feldelement                      *)
(*--------------------------------------------------------------------------*)

procedure FELDZUGRIFF;
  var ADDR,FELDADR,INDEX: OPERAND;
  begin
    if KELLER[TOP-1].ART<>FELD then
      SEM_ERROR('Indizierung einer einf. Variablen');
    ADDR.REG:=FREIES_INDEX_REGISTER;   (* freies Adressregister ermitteln   *)
    POP(INDEX);                        (* Feldindex und                     *)
    POP(FELDADR);                      (* Feldname aus Keller holen         *)
    ADDR.TYP:=INTEGER_SYM;             (* Der Operand ADDR wird vorerst     *)
                                       (* als normales INTEGER-Register be- *)
                                       (* trachtet, um den Index addieren   *)
                                       (* zu knnen                         *)
    ADDR.ART:=REGISTER;
                                       (* erzeuge einen Befehl, der die     *)
                                       (* Adresse des 0.ten Feldelementes   *)
                                       (* in ein Register ldt              *)
    WRITELN(CODEFILE,'LEA  ':17,REGISTER_NAME[INTEGER_SYM,ADDR.REG],
                              ',_',NB[FELDADR.NUMMER].NAME (*,'-',
                              ORD(FELDADR.TYP=INTEGER_SYM)+1*) );
    BEFEHL('ADD',ADDR,INDEX);          (* addiere Index zum Register        *)
    if FELDADR.TYP=INTEGER_SYM then    (* bei INTEGER-Gren 2 mal          *)
      BEFEHL('ADD',ADDR,INDEX);
    if INDEX.TYP <> INTEGER_SYM then
      TYP_FEHLER('Feldindex');
    ADDR.TYP:=FELDADR.TYP;             (* ab jetzt wird ADDR als Adresse    *)
                                       (* eines Feldelementes betrachtet    *)
    ADDR.ART:=ADR_REG;
    PUSH(ADDR);
  end;

(*--------------------------------------------------------------------------*)
(*            Bearbeitung fuer Name                                         *)
(*--------------------------------------------------------------------------*)

procedure NAMENS_BEARBEITUNG;
  var VARIABLE: OPERAND;
  begin
      if DEKLARATIONSTEIL then               (* Deklarationsteil            *)
        begin
          NB_INDEX:=ATTR;                    (* Test, ob bereits deklariert *)
          if NB[ATTR].DIMENSION<>(-1) then
            SEM_ERROR(CONCAT(NB[ATTR].NAME,' mehrfach deklariert'));
          NB[ATTR].DIMENSION:=0;             (* Annahme: kein Feld          *)
        end
      else                                   (* Anweisungsteil              *)
        begin
          VARIABLE.ART:=VARIAB;              (* Kennung VARIABLE und Nummer *)
          VARIABLE.NUMMER:=ATTR;             (* der Variablen festlegen     *)
          if NB[ATTR].DIMENSION=(-1) then
            SEM_ERROR(CONCAT(NB[ATTR].NAME,' nicht deklariert'));
          if NB[ATTR].DIMENSION>0 then       (* wenn Feldname               *)
            VARIABLE.ART:=FELD;
          VARIABLE.TYP:=NB[ATTR].TYP;
          PUSH(VARIABLE);                    (* Name kellern                *)
        end
  end;

(*--------------------------------------------------------------------------*)
(*            Bearbeitung fuer Konstante                                    *)
(*--------------------------------------------------------------------------*)

procedure ZAHL_BEARBEITUNG;
  var KONSTANTE: OPERAND;
  begin
    if DEKLARATIONSTEIL then      (* eine Konstante im Deklarationsteil ist *)
                                  (* die Feldlaenge des zuvor eingelesenen  *)
      begin                       (* Namens                                 *)
        NB[NB_INDEX].DIMENSION:=ATTR
      end
    else
      begin                       (* eine Konstante im Ausfhrungsteil mu  *)
        KONSTANTE.ART:=KONST;     (* als Operand gekellert werden           *)
        KONSTANTE.NUMMER:=ATTR;
        KONSTANTE.TYP:=INTEGER_SYM;
        PUSH(KONSTANTE);          (* INTEGER-Konstante kellern              *)
      end
  end;

procedure DEKLARATIONEN_AUSGEBEN;
  var l: integer;
  begin
    for NB_INDEX:=2 to LETZTER_NAME do
      with NB[NB_INDEX] do
        begin
          WRITE(CODEFILE,'_',NAME);
          (* Berechnung des Speicherbedarfs der Variablen *)
          if DIMENSION=0 then      (* einfache Variable           *)
            l:=1
          else
            l:=DIMENSION;          (* Feld                        *)
          l:=(ORD(TYP=INTEGER_SYM)+1)*l;
          WRITELN(CODEFILE,'DB':13-LENGTH(NAME),'   ',
                  l,' DUP(?)');
        end;
    WRITELN(CODEFILE);
    WRITE(CODEFILE,'PUBLIC ':19);
    for NB_INDEX:=2 to LETZTER_NAME do
      WRITE(CODEFILE,'_',NB[NB_INDEX].NAME,',');
    WRITELN(CODEFILE);
    WRITELN(CODEFILE,'_DATA	    ends');
    WRITELN(CODEFILE);
    WRITELN(CODEFILE,'_CODE       segment use16');
    WRITELN(CODEFILE,'            assume  cs:_CODE,ds:_DATA,ss:_STACK');
    WRITELN(CODEFILE);
    WRITELN(CODEFILE,'_',PROGRAMMNAME,' PROC NEAR':19-LENGTH(PROGRAMMNAME));
    DEKLARATIONSTEIL:=FALSE;
  end;

procedure PROGRAMM_ENDE;
  begin
    WRITELN;
    WRITELN('Ende der Uebersetzung');
    WRITELN(CODEFILE,'RET':15);
    WRITELN(CODEFILE,'_',PROGRAMMNAME,'ENDP':15-LENGTH(PROGRAMMNAME));
    {WRITELN(CODEFILE,'PUBLIC      _',PROGRAMMNAME);}
    WRITELN(CODEFILE,'start:');
    WRITELN(CODEFILE,'            MOV  AX, SEG _DATA');
    WRITELN(CODEFILE,'            MOV  DS,AX');
    WRITELN(CODEFILE,'            CALL _',PROGRAMMNAME);
    WRITELN(CODEFILE,'            MOV  AX,4C00h');
    WRITELN(CODEFILE,'            INT  21h');
    WRITELN(CODEFILE,'_CODE       ends');
    WRITELN(CODEFILE,'_STACK      segment para stack');
    WRITELN(CODEFILE,'            db 800 dup (?)');
    WRITELN(CODEFILE,'_STACK      ends');
    WRITELN(CODEFILE,'     	      end   start');
    CLOSE(CODEFILE);
  end;

(*--------------------------------------------------------------------------*)

  begin (* SEMANTISCHE_ANALYSE_UND_CODEERZEUGUNG *)
     case SYMBL of
       ZAHL:                               ZAHL_BEARBEITUNG;
       NAME:                               NAMENS_BEARBEITUNG;
       OR_SYM,AND_SYM:                     DYADISCHER_OPERATOR(BOOLEAN_SYM);
       MINUS,PLUS,MAL,GETEILT:             DYADISCHER_OPERATOR(INTEGER_SYM);
       KLEINER,GROESSER,GLEICH,UNGLEICH:   VERGLEICHSOPERATOR;
       DOPPGLEICH:                         ZUW_OPERATOR;
       BEGIN_SYM:                          DEKLARATIONEN_AUSGEBEN;
       END_SYM:                            PROGRAMM_ENDE;
       INTEGER_SYM,BOOLEAN_SYM:            NB[NB_INDEX].TYP:=SYMBL;
       IF_SYM,THEN_SYM,ELSE_SYM,END_IF:    IF_BEARBEITUNG;
       WHILE_SYM,DO_SYM,END_WHILE:         WHILE_BEARBEITUNG;
       NEGATIV:                            KOMPLEMENTIEREN;
       INDEX_SYM:                          FELDZUGRIFF;
       else SEM_ERROR('Fehlerhaftes Terminalsymbol')
     end
  end;

begin
    FREIE_REGISTER :=[AX..DI];
    MARKENZAEHLER:=1;         (* Markenzaehler initialisieren          *)
    DEKLARATIONSTEIL:=TRUE;
    TOP:=0;                   (* Keller initialisieren                 *)
    ASSIGN(CODEFILE,CONCAT(DATEINAME,'.asm'));
    REWRITE(CODEFILE);
    WRITELN(CODEFILE,'            .386p');
    WRITELN(CODEFILE,'_CODE       segment use16');
    WRITELN(CODEFILE,'            assume cs:_CODE, ds:_DATA, ss:_STACK');
    WRITELN(CODEFILE,'_CODE       ends');
    WRITELN(CODEFILE,'_DATA       segment use16');
    WRITELN(CODEFILE);
end.
