(*==========================================================================*)
(* Muster-Compiler fuer eine PASCAL-aehnliche Trivialsprache                *)
(*                                                                          *)
(* Parser                                                                   *)
(*                                                                          *)
(* Lexikalische Analyse wird fuer jedes Token vom Parser aufgerufen.        *)
(* Semantische Analyse bekommt jeses Token einzeln vom Parser ber die      *)
(* Prozedur UPN in umgekehrter polnischer Notation.                         *)
(* Codeerzeugung ist Bestandteil der Semantischen Analyse.                  *)
(*==========================================================================*)

program MINI_PASCAL_COMPILER;

  uses SCANNER,CODER;

(*==========================================================================*)
(*     Syntaxanalyse nach der Methode des rekursiven Abstiegs               *)
(*==========================================================================*)


  procedure DEKLARATIONEN;                                         forward;
  procedure VEREINBARUNG;                                          forward;
  procedure ANWEISUNG;                                             forward;
  procedure AUSDRUCK;                                              forward;
  procedure VARIABLE;                                              forward;
  procedure FAKTOR;                                                forward;
  procedure SUMMAND;                                               forward;
  procedure KOMPONENTE;                                            forward;
  procedure BLOCK;                                                 forward;
  procedure VERZWEIGUNG;                                           forward;
  procedure SCHLEIFE;                                              forward;
  procedure ZUWEISUNG;                                             forward;
  procedure FEHLER(MELDUNG: STRING);                               forward;
  procedure TEST(ERWARTET: SYMBOL);                                forward;
  function  AKTUELLES_TOKEN(ERWARTET: SYMBOL): BOOLEAN;            forward;
(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* programm : 'PROGRAM' NAME ';' deklarationen block ; *)
(* UPN      : deklarationen 'BEGIN' block 'END' ;      *)

  procedure PROGRAMM;
    begin
      TEST(PROGRAM_SYM);                (* = PROGRAM ?                      *)
      PROGRAMMNAME:=NB[TOKEN.ATTR].NAME;(* Programmname bereits gelesen     *)
      TEST(NAME);                       (* ist es wirklich ein Name ?       *)
      TEST(SEMIKOLON);                  (* nchstes Token = ';' ?           *)
      DEKLARATIONEN;
      UPN(BEGIN_SYM,0);                 (* BEGIN ausgeben                   *)
      BLOCK;
      UPN(END_SYM,0);                   (* END ausgeben                     *)
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* deklarationen : 'VAR' vereinbarung... ; *)
(* UPN           : vereinbarung...       ; *)

  procedure DEKLARATIONEN;
    begin
        TEST(VAR_SYM);             (* Pruefung, ob aktuelles TOKEN = VAR_SYM
                                      wenn ja   -> naechstes TOKEN lesen
                                      wenn nein -> Meldung: 'VAR' erwartet
                                   *)
        repeat
          VEREINBARUNG
        until TOKEN.KEY <> NAME;   (* da Vereinbarung mit NAME beginnt      *)
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* vereinbarung : NAME [ '[' ZAHL ']' ] ':' { 'INTEGER' | 'BOOLEAN' } ';' ; *)
(* UPN          : NAME [ ZAHL ] { 'INTEGER' | 'BOOLEAN' } ;                  *)

  procedure VEREINBARUNG;
    begin
      with TOKEN do
        begin
          UPN(NAME,ATTR);                 (* Name ausgeben                  *)
          TEST(NAME);                     (* Test auf NAME, naechstes TOKEN *)
          if AKTUELLES_TOKEN(ECK_AUF) then(* Test auf [, naechstes TOKEN    *)
            begin
              UPN(ZAHL,ATTR);             (* Dimension ausgeben             *)
              TEST(ZAHL);                 (* Test auf ZAHL, naechstes TOKEN *)
              TEST(ECK_ZU);               (* Test auf ], naechstes TOKEN    *)
            end;
          TEST(DOPPELP);                  (* Test auf :, naechstes TOKEN    *)
          if not (TOKEN.KEY in [INTEGER_SYM,BOOLEAN_SYM]) then
            FEHLER('INTEGER oder BOOLEAN')
          else
            begin
              UPN(TOKEN.KEY,0);           (* Typ ausgeben                   *)
              LEXIKALISCHE_ANALYSE;
            end;
          TEST(SEMIKOLON);                (* Test auf ;, naechstes TOKEN    *)
      end
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* anweisung : verzweigung | schleife | block | zuweisung ; *)
(* UPN       :                 -"-                          *)

  procedure ANWEISUNG;
    begin
        case TOKEN.KEY of
          IF_SYM     : VERZWEIGUNG;
          WHILE_SYM  : SCHLEIFE;
          BEGIN_SYM  : BLOCK;
          else         if TOKEN.KEY=NAME then
                         ZUWEISUNG
        end;
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* ausdruck : komponente [ { '>' | '<' | '=' } komponente ] ; *)
(* UPN      : komponente [ komponente { '>' | '<' | '=' } ] ; *)

  procedure AUSDRUCK;
    var _OPERATOR: SYMBOL;
      begin
        KOMPONENTE;                  (* erste Komponente des Ausdruckes     *)
        _OPERATOR:=TOKEN.KEY;         (* etwaigen Operator zwischenspeichern *)
        if _OPERATOR in [GROESSER,KLEINER,GLEICH,UNGLEICH] then
          begin
            LEXIKALISCHE_ANALYSE;
            KOMPONENTE;              (* weitere Ausdrucks-Komponenten       *)
            UPN(_OPERATOR,0);         (* Operator ausgeben (UPN)             *)
          end
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* komponente : ['-'] summand [ { '+' | '-' | 'OR' } summand ]... ; *)
(* UPN        : summand ['-'] [ summand { '+' | '-' | 'OR' } ]... ; *)

  procedure KOMPONENTE;
      var MON_MINUS: BOOLEAN;
          _OPERATOR:  SYMBOL;
    begin
        MON_MINUS:=AKTUELLES_TOKEN(MINUS);         (* negatives Vorzeichen ?*)
        SUMMAND;                                   (* erster Summand        *)
        if MON_MINUS then UPN(NEGATIV,0);          (* mon. Minus ausgeben   *)
        while TOKEN.KEY in [PLUS,MINUS,OR_SYM] do
          begin
            _OPERATOR:=TOKEN.KEY;                   (* Operator zw.speichern *)
            LEXIKALISCHE_ANALYSE;
            SUMMAND;                               (* weitere Summanden     *)
            UPN(_OPERATOR,0);                       (* Operator ausgeben     *)
          end
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* summand : faktor [ { '*' | '/' | 'AND' } faktor ]... ; *)
(* UPN:    : faktor [ faktor { '*' | '/' | 'AND' } ]... ; *)

  procedure SUMMAND;
      var _OPERATOR: SYMBOL;
    begin
        FAKTOR;                                    (* erster Faktor         *)
        while TOKEN.KEY in [MAL,GETEILT,AND_SYM] do
          begin
            _OPERATOR:=TOKEN.KEY;                   (* Operator zw.speichern *)
            LEXIKALISCHE_ANALYSE;
            FAKTOR;                                (* weitere Faktoren      *)
            UPN(_OPERATOR,0);                       (* Operator ausgeben     *)
          end
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* faktor : ZAHL | '(' ausdruck ')' | variable ; *)
(* UPN    : ZAHL | ausdruck | variable ;         *)

  procedure FAKTOR;
    begin
        if TOKEN.KEY=ZAHL then
          begin
            UPN(TOKEN.KEY,TOKEN.ATTR);             (* Zahl ausgeben         *)
            LEXIKALISCHE_ANALYSE
          end
        else
          if AKTUELLES_TOKEN(KL_AUF) then
            begin
              AUSDRUCK;                            (* geklammerter Ausdruck *)
              TEST(KL_ZU);
            end
          else VARIABLE
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* variable : NAME [ '[' ausdruck ']' ] ;   *)
(* UPN      : NAME [ ausdruck INDEX_SYM ] ; *)

  procedure VARIABLE;
    var NR: INTEGER;
    begin
        NR:=TOKEN.ATTR;
        TEST(NAME);
        UPN(NAME,NR);                         (* Name ausgeben              *)
        if AKTUELLES_TOKEN(ECK_AUF) then
          begin                               (* Index bearbeiten           *)
            AUSDRUCK;
            UPN(INDEX_SYM,0);                 (* Pseudo-TOKEN 'Feldzugriff' *)
            TEST(ECK_ZU);
          end
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* block : 'BEGIN' anweisung [ ';' anweisung ]... 'END' ; *)
(* UPN   : anweisung... ;                                 *)

  procedure BLOCK;                       (* BEGIN-END-Block                 *)
    begin
     TEST(BEGIN_SYM);
      repeat
        ANWEISUNG;
        if TOKEN.KEY=SEMIKOLON then      (* ';' mu zwischen Anweisungen !  *)
          LEXIKALISCHE_ANALYSE
        else
          if TOKEN.KEY<>END_SYM then
            FEHLER('";"');
      until TOKEN.KEY in [END_SYM,DATEI_ENDE];
      TEST(END_SYM);
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* verzweig.: 'IF' ausdruck 'THEN' anweisung [ 'ELSE' anweisung ] ;         *)
(* UPN      : 'IF' ausdruck 'THEN' anweisung [ 'ELSE' anweisung ] END_IF ;  *)

  procedure VERZWEIGUNG;
    begin
          UPN(TOKEN.KEY,0);                   (* IF ausgeben                *)
          LEXIKALISCHE_ANALYSE;
          AUSDRUCK;
          UPN(THEN_SYM,0);                    (* THEN ausgeben              *)
          TEST(THEN_SYM);
          ANWEISUNG;
          if AKTUELLES_TOKEN(ELSE_SYM) then
            begin
              UPN(ELSE_SYM,0);                (* ELSE ausgeben              *)
              ANWEISUNG
            end;
          UPN(END_IF,0);                      (* Pseudo-Token ENDIF ausgeben*)
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* schleife : 'WHILE' ausdruck 'DO' anweisung ;           *)
(* UPN      : 'WHILE' ausdruck 'DO' anweisung END_WHILE ; *)

  procedure SCHLEIFE;
    begin
          UPN(TOKEN.KEY,0);                  (* WHILE ausgeben    *)
          LEXIKALISCHE_ANALYSE;
          AUSDRUCK;
          TEST(DO_SYM);
          UPN(DO_SYM,0);                     (* DO ausgeben       *)
          ANWEISUNG;
          UPN(END_WHILE,0);                  (* ENDWHILE ausgeben *)
    end;

(*--------------------------------------------------------------------------*)
(*--------------------------------------------------------------------------*)

(* zuweisung : variable ':=' ausdruck ; *)
(* UPN       : variable ausdruck ':=' ; *)

  procedure ZUWEISUNG;
    begin
        VARIABLE;
        TEST(DOPPGLEICH);
        AUSDRUCK;
        UPN(DOPPGLEICH,0);                         (* := ausgeben *)
    end;

(*--------------------------------------------------------------------------*)
(*             Syntaxfehler melden und Programm abbrechen                   *)
(*--------------------------------------------------------------------------*)

  procedure FEHLER(MELDUNG: STRING);
    begin
      WRITELN(' << ',MELDUNG,' erwartet!'); HALT;
    end;

(*--------------------------------------------------------------------------*)
(*             Pruefung auf unbedingt erwartetes Terminalsymbol             *)
(*--------------------------------------------------------------------------*)

  procedure TEST(ERWARTET: SYMBOL);
    begin
      if TOKEN.KEY<>ERWARTET
      then FEHLER(LESBAR[ERWARTET])
      else LEXIKALISCHE_ANALYSE;(* wenn erkannt, dann naechstes TOKEN lesen *)
    end;

(*--------------------------------------------------------------------------*)
(*             Pruefung auf moeglicherweise erwartetes Terminalsymbol       *)
(*--------------------------------------------------------------------------*)

  function AKTUELLES_TOKEN(ERWARTET: SYMBOL): BOOLEAN;
    begin
      if TOKEN.KEY=ERWARTET then
        begin
          LEXIKALISCHE_ANALYSE;
          AKTUELLES_TOKEN:=TRUE
        end
      else
        AKTUELLES_TOKEN:=FALSE;
    end;

(*--------------------------------------------------------------------------*)

  begin (* Compiler *)
    LEXIKALISCHE_ANALYSE;     (* erstes TOKEN bestimmen                *)
    PROGRAMM;                 (* PROGRAMM analysieren                  *)

    WRITELN;
    CLOSE(QUELLE);
  end.
